package be.iminds.ilabt.jfed.experimenter_gui.bugreporting;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.conn.ssl.NoopHostnameVerifier;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.mime.MultipartEntityBuilder;
import org.apache.http.entity.mime.content.FileBody;
import org.apache.http.impl.client.HttpClients;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.net.ssl.SSLContext;
import java.io.*;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.util.concurrent.FutureTask;

/**
 * User: twalcari
 * Date: 7/17/2014
 * Time: 14:35
 */
public class BugReportToServerTask extends BugReportToFileTask {
    private static final Logger LOG = LoggerFactory.getLogger(BugReportToServerTask.class);

    private static final String BUG_REPORT_URL = "https://flsmonitor.fed4fire.eu/jfedexperimenter/submitv2.php";

    private static final SSLConnectionSocketFactory unsafeSslSocketFactory = createUnsafeSslConnectionSocketFactory();

    private static SSLConnectionSocketFactory createUnsafeSslConnectionSocketFactory() {
        try {
            SSLContext sslcontext = SSLContext.getInstance("TLS");
            sslcontext.init(null, null, null);
            return new SSLConnectionSocketFactory(sslcontext, NoopHostnameVerifier.INSTANCE);

        } catch (NoSuchAlgorithmException | KeyManagementException e) {
            throw new RuntimeException("Could not create an unsafe connection manager", e);
        }

    }

    public BugReportToServerTask(BugReportComposer bugReportComposer) {
        super(bugReportComposer, createTempFile());
    }

    private static File createTempFile() {
        try {
            File file = File.createTempFile("bug-report", ".xml");
            file.deleteOnExit();
            return file;
        } catch (IOException e) {
            LOG.error("Could not create temporary bug report file", e);
            throw new RuntimeException("Could not create temporary bug report file", e);
        }

    }

    @Override
    protected String call() throws Exception {


        super.call();

        try {
            final HttpClient httpClient = HttpClients.custom().setSSLSocketFactory(unsafeSslSocketFactory).build();

            final HttpPost httpPost = new HttpPost(BUG_REPORT_URL);
            MultipartEntityBuilder multipartEntityBuilder = MultipartEntityBuilder.create();
            FileBodyWithProgress fileBodyWithProgress = new FileBodyWithProgress(outputFile,
                    ContentType.APPLICATION_XML.getMimeType(), ContentType.APPLICATION_XML.getCharset().toString());
            multipartEntityBuilder.addPart("bugreport", fileBodyWithProgress);

            httpPost.setEntity(multipartEntityBuilder.build());

            FutureTask<HttpResponse> posterTask = new FutureTask<>(
                    () -> httpClient.execute(httpPost));

            Thread thread = new Thread(posterTask, "BugReportPoster");
            thread.start();

            while (!posterTask.isDone()) {
                updateToServerProgress(fileBodyWithProgress.getProgress(), 101);
                Thread.sleep(500);
            }

            updateMessage("Waiting for response from server");

            HttpResponse response = posterTask.get();
            HttpEntity entity = response.getEntity();
            if (entity != null) {
                StringBuilder result = new StringBuilder();
                try (BufferedReader reader = new BufferedReader(new InputStreamReader(entity.getContent(), "UTF-8"))) {
                    String line;
                    while ((line = reader.readLine()) != null) {
                        result.append(line);
                    }
                    return result.toString();
                } catch (IOException ex) {
                    LOG.error("Error while reading from bug report server", ex);
                    return null;
                }
            } else {
                LOG.error("bug report server did not return info. {}: {}",
                        response.getStatusLine().getStatusCode(), response.getStatusLine().getReasonPhrase());
                return null;
            }
        } finally {
            //noinspection ResultOfMethodCallIgnored
            outputFile.delete();
        }
    }


    private static final double PARENT_SHARE_IN_PROGRESS = 0.7;

    /**
     * we override this method because the parent-class only does part of the work. We take this into account here.
     *
     * @param workDone
     * @param totalWork
     */
    @Override
    protected void updateProgress(long workDone, long totalWork) {
        if (workDone == -1)
            super.updateProgress(-1, 0);
        else
            super.updateProgress((PARENT_SHARE_IN_PROGRESS * workDone) / totalWork, 1);
    }

    /**
     * We assume that the parent has finished here.
     *
     * @param workDone  Work done to upload to server
     * @param totalWork Total work to upload to server
     */
    private void updateToServerProgress(long workDone, long totalWork) {
        assert (workDone <= totalWork) : "Received work done of " + workDone + " while total work is " + totalWork;
        super.updateProgress(PARENT_SHARE_IN_PROGRESS + (1.0 - PARENT_SHARE_IN_PROGRESS) * ((1.0 * workDone) / totalWork), 1);
    }

    @Override
    protected void updateStartFileWriteMessage() {
        updateMessage("Preparing bug report");
    }

    @Override
    protected void updateEndFileWriteMessage() {
        updateMessage("Sending bug report to server");
    }

    private static class FileBodyWithProgress extends FileBody {
        private OutputStreamProgress outstream;


        private FileBodyWithProgress(File file, String mimeType, String charset) {
            super(file, mimeType, charset);
        }

        @Override
        public void writeTo(OutputStream outstream) throws IOException {
            this.outstream = new OutputStreamProgress(outstream);
            super.writeTo(this.outstream);
        }

        /**
         * Progress: 0-100
         */
        public int getProgress() {
            if (outstream == null) {
                return 0;
            }
            long contentLength = getContentLength();
            if (contentLength <= 0) { // Prevent division by zero and negative values
                return 0;
            }
            long writtenLength = outstream.getWrittenLength();
            return (int) (100 * writtenLength / contentLength);
        }
    }

    private static class OutputStreamProgress extends OutputStream {

        private final OutputStream outstream;
        private volatile long bytesWritten = 0;

        public OutputStreamProgress(OutputStream outstream) {
            this.outstream = outstream;
        }

        @Override
        public void write(int b) throws IOException {
            outstream.write(b);
            bytesWritten++;
        }

        @Override
        public void write(byte[] b) throws IOException {
            outstream.write(b);
            bytesWritten += b.length;
        }

        @Override
        public void write(byte[] b, int off, int len) throws IOException {
            outstream.write(b, off, len);
            bytesWritten += len;
        }

        @Override
        public void flush() throws IOException {
            outstream.flush();
        }

        @Override
        public void close() throws IOException {
            outstream.close();
        }

        public long getWrittenLength() {
            return bytesWritten;
        }
    }

    @Override
    protected void failed() {
        LOG.error("Writing bugreport to server failed: {}",
                getException() != null ? getException().getMessage() : "no error",
                getException());

    }
}
